{*
 *  LDLITE, a program for viewing *.dat files.
 *  Copyright (C) 1998  Paul J. Gyugyi
 *
 *  Translated to Pascal (Delphi 2) by Anders Isaksson 1998
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*}

unit ldliteVRh;

interface

uses
  UniDib
  ;

const
  MAXSIZE =  786432;

type

  MyReal  = double;

  PInteger  = ^integer;
  IntArr    = array[0..MAXSIZE] of integer;
  ByteArr   = array[0..MAXSIZE] of byte;
  PIntArr   = ^IntArr;
  PByteArr  = ^ByteArr;

  Vector3d = record
    x, y, z: MyReal;
  end;

  PVector3d = ^Vector3d;

  Matrix3d = record
    a, b, c, d, e, f, g, h, i:  MyReal;
  end;

  PMatrix3d = ^Matrix3d;

  ZIMAGE = record
    zbuffer:  PIntArr;
(*
    r, g, b:  PByteArr;
*)
    DIB:      TUniDIB;
    rows,
    cols,
	  dirty_x1,
	  dirty_x2,
	  dirty_y1,
	  dirty_y2: integer;
  end;

  ZPOINT = record
    x, y, z: integer;
  end;
  PZPOINT = ^ZPOINT;

  ZPLANE = record
    A, B, C, D: MyReal;
  end;

  ZCOLOR  = record
    r, g, b, a: byte; // red, green, blue, alpha 255 == solid, 0 = transparent
  end;
  PZCOLOR = ^ZCOLOR;
const

  MAX_INCLUDE_DEPTH = 64;


function max(const a, b: integer): integer;
function min(const a, b: integer): integer;
function minmax(const a, b, d: integer): integer;
function maxmin(const a, b, d: integer): integer;
function max3(const a, b, c: integer): integer;
function min3(const a, b, c: integer): integer;
function minmax3(const a, b, c, d: integer): integer;
function maxmin3(const a, b, c, d: integer): integer;

implementation

function max(const a, b: integer): integer;
begin
  if a > b then Result := a else Result := b;
end;

function min(const a, b: integer): integer;
begin
  if a < b then Result := a else Result := b;
end;

function minmax(const a, b, d: integer): integer;
begin
  if a > b then Result := a else Result := b;
  if Result > d then Result := d;
end;

function maxmin(const a, b, d: integer): integer;
begin
  if a < b then Result := a else Result := b;
  if Result < d then Result := d;
end;

function max3(const a, b, c: integer): integer;
begin
  if a > b then
    if a > c then Result := a else Result := c
  else
    if b > c then Result := b else Result := c
end;

function min3(const a, b, c: integer): integer;
begin
  if a < b then
    if a < c then Result := a else Result := c
  else
    if b < c then Result := b else Result := c
end;

function minmax3(const a, b, c, d: integer): integer;
begin
  if a > b then
    if a > c then Result := a else Result := c
  else
    if b > c then Result := b else Result := c;
  if Result > d then Result := d;
end;

function maxmin3(const a, b, c, d: integer): integer;
begin
  if a < b then
    if a < c then Result := a else Result := c
  else
    if b < c then Result := b else Result := c;
  if Result < d then Result := d;
end;


end.

