{*
 *  LDLITE, a program for viewing *.dat files.
 *  Copyright (C) 1998  Paul J. Gyugyi
 *
 *  Translated to Pascal (Delphi 2) by Anders Isaksson 1998
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*}

unit stub;

interface

uses
  ldliteVRh, Main, UniDIB
  ;

const
  INT_MAX = maxint;
  INT_MIN = -maxint;

var
  z:            ZIMAGE;
  zShading:     boolean = True;
  zDetailLevel: integer;

function zGetColsize: integer;
function zGetRowsize: integer;

procedure zAddLine(var zp: ZIMAGE; const pc1, pc2: PZPOINT;
                   const zcp, zcs: PZCOLOR; const replace: integer);
procedure zAddTriangle(var zp: ZIMAGE; const p1, p2, p3: PZPOINT;
                       const zcp, zcs: PZCOLOR);
procedure zDraw(const zDC: pointer);
procedure zFree;
function  zReset(rows, cols: integer): integer;
function  edge_color(const c: integer): integer;
procedure zStep(const step, pause: integer);
procedure render_line(const vp1, vp2: vector3d; const c: integer);
procedure render_triangle(const vp1, vp2, vp3: vector3d; const c: integer);
procedure render_quad(const vp1, vp2, vp3, vp4: vector3d; const c: integer);
procedure render_five(const vp1, vp2, vp3, vp4: vector3d; const c: integer);

implementation

uses
  SysUtils,
  ytab, Debug
  ;

//

function Trunc(x: MyReal): integer;
begin
  Result := Round(x);
end;

// Returns ldraw edge (i.e. contrasting) color.
// Note that complementing is not commutative.
// c != complement_color(complement_color(c)), for example when c = 7

function edge_color(const c: integer): integer;
begin
	case c of
	0: Result := 8;
	1: Result := 9;
	2: Result := 10;
	3: Result := 11;
	4: Result := 12;
	5: Result := 13;
	6: Result := 0;
	7: Result := 8;
	8: Result := 0;
	9: Result := 1;
	10: Result := 2;
	11: Result := 3;
	12: Result := 4;
	13: Result := 5;
	14: Result := 8;
	15: Result := 8;
	17: Result := 2;
	18: Result := 14;
	19: Result := 6;
	20: Result := 8;
	21: Result := 0;
	22: Result := 8;
	23: Result := 8;
	25: Result := 8;
	33: Result := 9;
	34: Result := 10;
	36: Result := 12;
  39: Result := 8;
	41: Result := 1;
	42: Result := 2;
	46: Result := 8;
	47: Result := 8;
	57: Result := 8;
	382: Result := 6;
	431: Result := 2;
	495: Result := 14;
	334: Result := 8;
	383: Result := 8;
	else begin
      if (c > 256) then begin
        // dithered colors get complement of higher color
        Result := edge_color( ((c-256) div 16) );
      end
      else begin
        // complement by flipping bit of base color
        Result := ((c and $0f) xor $08);
      end
    end
	end;
end;

//
// zcp is the primary color
// zcs is the secondary color used for dithering
procedure translate_color(const c: integer; var zcp, zcs: ZCOLOR);
var
  basecolor:  integer;
  highcolor:  integer;
  unused:     ZCOLOR;
begin
	case c of
  	// color bits:  Transparency Intensity Red Green Blue
		//basic colors:
	 0: begin
		zcp.r := $033; zcp.g := $033; zcp.b := $033; zcp.a := $0ff;
		zcs.r := $033; zcs.g := $033; zcs.b := $033; zcs.a := $0ff;
    end;
	 1: begin
		zcp.r := $000; zcp.g := $033; zcp.b := $0b2; zcp.a := $0ff;
		zcs.r := $000; zcs.g := $033; zcs.b := $0b2; zcs.a := $0ff;
    end;
	 2: begin
		zcp.r := $000; zcp.g := $07f; zcp.b := $033; zcp.a := $0ff;
		zcs.r := $000; zcs.g := $07f; zcs.b := $033; zcs.a := $0ff;
    end;
	 3: begin
		zcp.r := $000; zcp.g := $0aa; zcp.b := $0aa; zcp.a := $0ff;
		zcs.r := $000; zcs.g := $0aa; zcs.b := $0aa; zcs.a := $0ff;
    end;
	 4: begin
		zcp.r := $0cc; zcp.g := $000; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0cc; zcs.g := $000; zcs.b := $000; zcs.a := $0ff;
    end;
	 5: begin
		zcp.r := $0ff; zcp.g := $033; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $033; zcs.b := $099; zcs.a := $0ff;
    end;
	 6: begin
		zcp.r := $066; zcp.g := $033; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $066; zcs.g := $033; zcs.b := $000; zcs.a := $0ff;
    end;
	 7: begin
		zcp.r := $0aa; zcp.g := $0aa; zcp.b := $0aa; zcp.a := $0ff;
		zcs.r := $0aa; zcs.g := $0aa; zcs.b := $0aa; zcs.a := $0ff;
    end;
	 8: begin
		zcp.r := $066; zcp.g := $066; zcp.b := $058; zcp.a := $0ff;
		zcs.r := $066; zcs.g := $066; zcs.b := $058; zcs.a := $0ff;
    end;
	 9: begin
		zcp.r := $000; zcp.g := $080; zcp.b := $0ff; zcp.a := $0ff;
		zcs.r := $000; zcs.g := $080; zcs.b := $0ff; zcs.a := $0ff;
    end;
	10: begin
		zcp.r := $033; zcp.g := $0ff; zcp.b := $066; zcp.a := $0ff;
		zcs.r := $033; zcs.g := $0ff; zcs.b := $066; zcs.a := $0ff;
    end;
	11: begin
		zcp.r := $055; zcp.g := $0aa; zcp.b := $0ff; zcp.a := $0ff;
		zcs.r := $055; zcs.g := $0aa; zcs.b := $0ff; zcs.a := $0ff;
    end;
	12: begin
		zcp.r := $0ff; zcp.g := $055; zcp.b := $055; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $055; zcs.b := $055; zcs.a := $0ff;
    end;
	13: begin
		zcp.r := $0ff; zcp.g := $0b0; zcp.b := $0cc; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0b0; zcs.b := $0cc; zcs.a := $0ff;
    end;
	14: begin
		zcp.r := $0ff; zcp.g := $0e5; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0e5; zcs.b := $000; zcs.a := $0ff;
    end;
	15: begin
		zcp.r := $0ff; zcp.g := $0ff; zcp.b := $0ff; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $0ff;
    end;
	17: begin
		zcp.r := $066; zcp.g := $0f0; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $066; zcs.g := $0f0; zcs.b := $099; zcs.a := $0ff;
    end;
	18: begin
		zcp.r := $0ff; zcp.g := $0ff; zcp.b := $080; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $080; zcs.a := $0ff;
    end;
	19: begin
		zcp.r := $0cc; zcp.g := $0aa; zcp.b := $066; zcp.a := $0ff;
		zcs.r := $0cc; zcs.g := $0aa; zcs.b := $066; zcs.a := $0ff;
    end;
	20: begin
		zcp.r := $0e0; zcp.g := $0cc; zcp.b := $0f0; zcp.a := $0ff;
		zcs.r := $0e0; zcs.g := $0cc; zcs.b := $0f0; zcs.a := $0ff;
    end;
	21: begin
		zcp.r := $0e0; zcp.g := $0ff; zcp.b := $0b0; zcp.a := $0ff;
		zcs.r := $0e0; zcs.g := $0ff; zcs.b := $0b0; zcs.a := $0ff;
    end;
	22: begin
		zcp.r := $099; zcp.g := $033; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $099; zcs.g := $033; zcs.b := $099; zcs.a := $0ff;
    end;
	23: begin
		zcp.r := $04c; zcp.g := $000; zcp.b := $0cc; zcp.a := $0ff;
		zcs.r := $04c; zcs.g := $000; zcs.b := $0cc; zcs.a := $0ff;
    end;
	25: begin
		zcp.r := $0ff; zcp.g := $066; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $066; zcs.b := $000; zcs.a := $0ff;
    end;
		// transparent colors:
	33: begin
		zcp.r := $000; zcp.g := $000; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	34: begin
		zcp.r := $000; zcp.g := $050; zcp.b := $018; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	36: begin
		zcp.r := $0cc; zcp.g := $000; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	39: begin
		zcp.r := $099; zcp.g := $099; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	41: begin
		zcp.r := $099; zcp.g := $0c0; zcp.b := $0f0; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	42: begin
		zcp.r := $0cc; zcp.g := $0ff; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	46: begin
		zcp.r := $0f0; zcp.g := $0c4; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	47: begin
		zcp.r := $0ff; zcp.g := $0ff; zcp.b := $0ff; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	57: begin
		zcp.r := $0ff; zcp.g := $066; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $0ff; zcs.a := $000;
    end;
	382: begin
		zcp.r := $0cc; zcp.g := $0aa; zcp.b := $066; zcp.a := $0ff;
		zcs.r := $0cc; zcs.g := $0aa; zcs.b := $066; zcs.a := $0ff;
    end;
	431: begin
		zcp.r := $066; zcp.g := $0f0; zcp.b := $099; zcp.a := $0ff;
		zcs.r := $066; zcs.g := $0f0; zcs.b := $099; zcs.a := $0ff;
    end;
	495: begin
		zcp.r := $0ff; zcp.g := $0ff; zcp.b := $080; zcp.a := $0ff;
		zcs.r := $0ff; zcs.g := $0ff; zcs.b := $080; zcs.a := $0ff;
    end;
	334: begin
		zcp.r := $0ff; zcp.g := $0e5; zcp.b := $000; zcp.a := $0ff;
		zcs.r := $0cc; zcs.g := $000; zcs.b := $000; zcs.a := $0ff;
    end;
	383: begin
		zcp.r := $0ff; zcp.g := $0ff; zcp.b := $0ff; zcp.a := $0ff;
		zcs.r := $099; zcs.g := $099; zcs.b := $099; zcs.a := $0ff;
    end;
	else begin
      if (c > 256 ) then begin
        // Dithered colors
        basecolor := (c-256) div 16;
        highcolor := c and $0f;

        translate_color(basecolor, zcp, unused);
        translate_color(highcolor, unused, zcs);
      end
      else if ((c >= 32) and (c < 48)) then begin
        // Transparent colors
        basecolor := c - 32;
        translate_color(basecolor, zcp, unused);
        zcs.r := $00; zcs.g := $00; zcs.b := $00; zcs.a := $00;
      end
      else begin
        // unknown color
        zcp.r := $0ff; zcp.g := $0ff; zcp.b := $0ff; zcp.a := $0ff;
        zcs.r := $000; zcs.g := $000; zcs.b := $000; zcs.a := $0ff;
      end
    end;
	end;
end;

procedure render_line(const vp1, vp2: vector3d; const c: integer);
var
	p1, p2: ZPOINT;
	zc, zs: ZCOLOR;
begin
	p1.x := Trunc(vp1.x);
  p1.y := Trunc(vp1.y);
  p1.z := Trunc(vp1.z);
	p2.x := Trunc(vp2.x);
  p2.y := Trunc(vp2.y);
  p2.z := Trunc(vp2.z);
	translate_color(c,zc,zs);
	zAddLine(z, @p1, @p2, @zc, @zs, 0);
end;

procedure render_triangle(const vp1, vp2, vp3: vector3d; const c: integer);
var
	p1, p2, p3: ZPOINT;
	zc, zs: ZCOLOR;
begin
	p1.x := Trunc(vp1.x);
  p1.y := Trunc(vp1.y);
  p1.z := Trunc(vp1.z);
	p2.x := Trunc(vp2.x);
  p2.y := Trunc(vp2.y);
  p2.z := Trunc(vp2.z);
	p3.x := Trunc(vp3.x);
  p3.y := Trunc(vp3.y);
  p3.z := Trunc(vp3.z);
	translate_color(c,zc,zs);
	zAddTriangle(z, @p1, @p2, @p3, @zc, @zs);
end;

function above_line(const vp1, vp2, vp3: vector3d): integer;
var
	yslope: MyReal;
	y:      MyReal;
begin
	if (vp2.x = vp1.x) then begin
    Result := 1;
  end
  else begin
    yslope := (vp2.y - vp1.y) / (vp2.x-vp1.x);
    y := (yslope * (vp3.x-vp1.x)) + vp1.y;
    if (vp3.y > y) then begin
      Result := 1;
    end
    else begin
      Result := 0;
    end
  end
end;

procedure render_quad(const vp1, vp2, vp3, vp4: vector3d; const c: integer);
var
	p1, p2, p3: ZPOINT;
	zc, zs: ZCOLOR;
	bowtie_test_1:  integer;
	bowtie_test_2:  integer;
begin
	translate_color(c,zc,zs);
	// draw triangle vp1 vp2 vp3
	p1.x := Trunc(vp1.x);
  p1.y := Trunc(vp1.y);
  p1.z := Trunc(vp1.z);
	p2.x := Trunc(vp2.x);
  p2.y := Trunc(vp2.y);
  p2.z := Trunc(vp2.z);
	p3.x := Trunc(vp3.x);
  p3.y := Trunc(vp3.y);
  p3.z := Trunc(vp3.z);
	zAddTriangle(z, @p1, @p2, @p3, @zc, @zs);
	// draw triangle vp1 vp3 vp4
	p1.x := Trunc(vp1.x);
  p1.y := Trunc(vp1.y);
  p1.z := Trunc(vp1.z);
	p2.x := Trunc(vp3.x);
  p2.y := Trunc(vp3.y);
  p2.z := Trunc(vp3.z);
	p3.x := Trunc(vp4.x);
  p3.y := Trunc(vp4.y);
  p3.z := Trunc(vp4.z);
	zAddTriangle(z, @p1, @p2, @p3, @zc, @zs);
	// handle concave quads
	bowtie_test_1 := above_line(vp1, vp3, vp2);
	bowtie_test_2 := above_line(vp1, vp3, vp4);
	if (bowtie_test_1 = bowtie_test_2) then begin
		// vp2 and vp4 lie on same side of line segment.
		// quad probably has a slice missing.
		// drawing triangle vp2 vp3 vp4 will fill it in.
		p1.x := Trunc(vp2.x);
    p1.y := Trunc(vp2.y);
    p1.z := Trunc(vp2.z);
		p2.x := Trunc(vp3.x);
    p2.y := Trunc(vp3.y);
    p2.z := Trunc(vp3.z);
		p3.x := Trunc(vp4.x);
    p3.y := Trunc(vp4.y);
    p3.z := Trunc(vp4.z);
		zAddTriangle(z, @p1, @p2, @p3, @zc, @zs);
	end
end;

procedure render_five(const vp1, vp2, vp3, vp4: vector3d; const c: integer);
var
	p1, p2: ZPOINT;
	zc, zs: ZCOLOR;
begin
	if ((vp1.x = vp2.x) and (vp1.y = vp2.y)) then Exit;
	if (vp1.x = vp2.x) then begin
		// vertical line
		if ((vp3.x > vp1.x) and (vp4.x < vp1.x)) then Exit;
		if ((vp3.x < vp1.x) and (vp4.x > vp1.x)) then Exit;
	end
  else if (vp1.y = vp2.y) then begin
		if ((vp3.y > vp1.y) and (vp4.y < vp1.y)) then Exit;
		if ((vp3.y < vp1.y) and (vp4.y > vp1.y)) then Exit;
	end
  else begin
		if (above_line(vp1, vp2, vp3) <> above_line(vp1,vp2,vp4)) then Exit;
	end;
	p1.x := Trunc(vp1.x);
  p1.y := Trunc(vp1.y);
  p1.z := Trunc((vp1.z) - 1);
	p2.x := Trunc(vp2.x);
  p2.y := Trunc(vp2.y);
  p2.z := Trunc((vp2.z) - 1);
	translate_color(c,zc, zs);
	zAddLine(z, @p1, @p2, @zc, @zs, 1);
end;

procedure calc_zplane(var zpl: ZPLANE; const p1, p2, p3: ZPOINT);
begin
  zpl.A := p1.y * (p2.z - p3.z) +
           p2.y * (p3.z - p1.z) +
           p3.y * (p1.z - p2.z);
  zpl.B := p1.z * (p2.x - p3.x) +
           p2.z * (p3.x - p1.x) +
           p3.z * (p1.x - p2.x);
  zpl.C := (-1) * (p1.x * (p2.y - p3.y) +
                   p2.x * (p3.y - p1.y) +
                   p3.x * (p1.y - p2.y));
	zpl.D := (-1) * p1.x * (p2.y * p3.z - p3.y * p2.z) -
                  p2.x * (p3.y * p1.z - p1.y * p3.z) -
            		  p3.x * (p1.y * p2.z - p2.y * p1.z);
end;

procedure calc_zplane_z(const zpl: ZPLANE; var p: ZPOINT);
begin
	if (zpl.C <> 0) then begin
		p.z := Trunc((zpl.D + zpl.A * p.x + zpl.B * p.y) / zpl.C);
	end
  else begin
		p.z := INT_MAX; // do not draw
	end;
end;

function init_zimage(var zp: ZIMAGE; rows, cols: integer): integer;
var
  i, j: integer;
begin
	if (zp.zbuffer <> nil) then FreeMem(zp.zbuffer);
	zp.zbuffer := AllocMem(rows * cols * sizeof(integer));
	if (zp.zbuffer = nil) then begin
		Result := -1;
	end;
(*
	if (zp.r <> nil) then FreeMem(zp.r);
	zp.r := AllocMem(rows*cols*sizeof(byte));
	if (zp.r = nil) then begin
		FreeMem(zp.zbuffer);
		Result := -1;
	end;
	if (zp.g <> nil) then FreeMem(zp.g);
	zp.g := AllocMem(rows*cols*sizeof(byte));
	if (zp.g = nil) then begin
		FreeMem(zp.r);
		FreeMem(zp.zbuffer);
		Result := -1;
	end;
	if (zp.b <> nil) then FreeMem(zp.b);
	zp.b := AllocMem(rows*cols*sizeof(byte));
	if (zp.b = nil) then begin
		FreeMem(zp.g);
		FreeMem(zp.r);
		FreeMem(zp.zbuffer);
		Result := -1;
	end;
*)
	zp.rows := rows;
	zp.cols := cols;

  if (Assigned(zp.DIB)) then begin
    zp.DIB.Free;
    zp.DIB := nil;
  end;
  zp.DIB := TUniDIB.Create(zp.Rows, zp.Cols, 24, SBU_NONE);

	for i := 0 to Pred(zp.cols) do begin
    zp.DIB.Seek(0, i);
		for j := 0 to Pred(zp.rows) do begin
			zp.zbuffer[i*zp.rows + j] := INT_MAX;
      zp.DIB.SetSeqPixel($00FFFFFF);
(*
			zp.r[i*zp.cols + j] := $ff;
			zp.g[i*zp.cols + j] := $ff;
			zp.b[i*zp.cols + j] := $ff;
*)
		end;
	end;
	zp.dirty_x1  := 0;
	zp.dirty_x2  := rows-1;
	zp.dirty_y1  := 0;
	zp.dirty_y2  := cols-1;
//	zp.extent_x1 := rows-1;
//	zp.extent_x2 := 0;
//	zp.extent_y1 := cols-1;
//	zp.extent_y2 := 0;
//  Debug.Print(Format('init_zimage exit: %d %d %d %d',
//      [zp.dirty_x1, zp.dirty_x2, zp.dirty_y1, zp.dirty_y2]));
	Result := 0;
end;

type
  PByte =       ^Byte;

var
  light:        ZPOINT = (x:1; y:1; z:1);

procedure zAddTriangle(var zp: ZIMAGE; const p1, p2, p3: PZPOINT; const zcp, zcs: PZCOLOR);
var
	miny:         integer;
	maxy:         integer;
	p:            ZPOINT;
	miny21:        integer;
	maxy21:        integer;
	miny22:        integer;
	maxy22:        integer;
	miny23:        integer;
	maxy23:        integer;
	slope1:        MyReal;
	intercept1:    MyReal;
	slope2:        MyReal;
	intercept2:    MyReal;
	slope3:        MyReal;
	intercept3:    MyReal;
	first_point,
  second_point: integer;
	intersection: integer;
	zpl:          ZPLANE;
	pixel:        integer;
	zc:           PZCOLOR;
	x1, x2, y1, y2: integer;
  f:            MyReal;
	t1,t2,t3,t4,t5,t6,t7,t8:  MyReal;
  mxpixel:      integer;
  ptr:          PByte;
begin
	if (zp.zbuffer = nil) then begin
		Exit;
	end;

  mxpixel := zp.Rows * zp.Cols;

	calc_zplane(zpl, p1^, p2^, p3^);

	maxy := max3(p1.y, p2.y, p3.y);
	miny := min3(p1.y, p2.y, p3.y);

	begin
		x1 := maxmin3(p1.x, p2.x, p3.x,0);
		x2 := minmax3(p1.x, p2.x, p3.x,(zp.rows-1));
		y1 := maxmin3(p1.y, p2.y, p3.y,0);
		y2 := minmax3(p1.y, p2.y, p3.y,(zp.cols-1));
		zp.dirty_x1 := min(x1,zp.dirty_x1);
		zp.dirty_x2 := max(x2,zp.dirty_x2);
		zp.dirty_y1 := min(y1,zp.dirty_y1);
		zp.dirty_y2 := max(y2,zp.dirty_y2);
//		zp.extent_x1 := min(x1,zp.extent_x1);
//		zp.extent_x2 := max(x2,zp.extent_x2);
//		zp.extent_y1 := min(y1,zp.extent_y1);
//		zp.extent_y2 := max(y2,zp.extent_y2);
//    Debug.Print(Format('add_triangle: %d %d %d %d',
//          [zp.dirty_x1, zp.dirty_x2, zp.dirty_y1, zp.dirty_y2]));
	end;

		miny21 := min(p1.y, p2.y);
		maxy21 := max(p1.y, p2.y);
			if(p1.x <> p2.x) then begin
				slope1 := (p2.y - p1.y) / (p2.x - p1.x);
				intercept1 := p1.y - slope1 * p1.x;
      end;
		miny22 := min(p2.y, p3.y);
		maxy22 := max(p2.y, p3.y);
			if(p2.x <> p3.x) then begin
				slope2 := (p3.y - p2.y) / (p3.x - p2.x);
				intercept2 := p2.y - slope2 * p2.x;
      end;
		miny23 := min(p3.y, p1.y);
		maxy23 := max(p3.y, p1.y);
			if(p3.x <> p1.x) then begin
				slope3 := (p1.y - p3.y) / (p1.x - p3.x);
				intercept3 := p3.y - slope3 * p3.x;
      end;

	p.y := miny;
  while p.y < maxy do begin
		first_point := INT_MAX;
		second_point := INT_MIN;
		// first edge
//		miny2 := min(p1.y, p2.y);
//		maxy2 := max(p1.y, p2.y);
		if ((miny21 <> maxy21) and (p.y >= miny21) and (p.y <= maxy21)) then begin
			// must intersect
			if(p1.x <> p2.x) then begin
//				slope := (p2.y - p1.y) / (p2.x - p1.x);
//				intercept := p1.y - slope * p1.x;
				intersection := Trunc((p.y - intercept1) / slope1);
			end
      else begin
				intersection := p1.x;
			end;
//			first_point := min(first_point,intersection);
//			second_point := max(second_point, intersection);
			if (first_point > intersection) then first_point := intersection;
			if (second_point < intersection) then second_point := intersection;
		end;
		// second edge
//		miny2 := min(p2.y, p3.y);
//		maxy2 := max(p2.y, p3.y);
		if ((miny22 <> maxy22) and (p.y >= miny22) and (p.y <= maxy22)) then begin
			// must intersect
			if(p2.x <> p3.x) then begin
//				slope := (p3.y - p2.y) / (p3.x - p2.x);
//				intercept := p2.y - slope * p2.x;
				intersection := Trunc((p.y - intercept2) / slope2);
			end
      else begin
				intersection := p2.x;
			end;
//			first_point := min(first_point,intersection);
//			second_point := max(second_point, intersection);
			if (first_point > intersection) then first_point := intersection;
			if (second_point < intersection) then second_point := intersection;
		end;
		// third edge
//		miny2 := min(p3.y, p1.y);
//		maxy2 := max(p3.y, p1.y);
		if ((miny23 <> maxy23) and (p.y >= miny23) and (p.y <= maxy23)) then begin
			// must intersect
			if(p3.x <> p1.x) then begin
//				slope := (p1.y - p3.y) / (p1.x - p3.x);
//				intercept := p3.y - slope * p3.x;
				intersection := Trunc((p.y - intercept3) / slope3);
			end
      else begin
				intersection := p3.x;
			end;
//			first_point := min(first_point,intersection);
//			second_point := max(second_point, intersection);
			if (first_point > intersection) then first_point := intersection;
			if (second_point < intersection) then second_point := intersection;
		end;
//		first_point := max(0,first_point);
//		second_point := min(((zp.rows)-1), second_point);
  	if (first_point < 0) then first_point := 0;
		if (second_point > (zp.rows-1)) then second_point := zp.rows-1;
		p.x := first_point;
    while p.x <= second_point do begin
//			calc_zplane_z(zpl, p);
        if (zpl.C <> 0) then begin
          p.z := Trunc((zpl.D + zpl.A * p.x + zpl.B * p.y) / zpl.C);
        end
        else begin
          p.z := INT_MAX; // do not draw
        end;
// 			if((p.y >= 0) and (p.y < zp.cols)) then begin
				pixel := p.x + (zp.cols-1-p.y)*zp.rows {p.x+p.y*zp.rows{p.x*zp.cols+p.y};
				if((pixel>=0) and (pixel < mxpixel{(zp.rows * zp.cols)})) then begin
					if (zp.zbuffer[pixel] > p.z) then begin
						// dither colors
						if ((Odd(p.x+p.y)){ mod 2) <> 0}) then begin
							zc := zcp;
						end
            else begin
							zc := zcs;
						end;
						if (zc.a > 0) then begin
							zp.zbuffer[pixel] := p.z;
(*
							zp.r[pixel]:= zc.r;
							zp.g[pixel]:= zc.g;
							zp.b[pixel]:= zc.b;
*)
              if (zShading) then begin
                f := (light.x * zpl.A + light.y *zpl.B - light.z*zpl.C) /
                  sqrt(light.x*light.x + light.y*light.y + light.z*light.z) /
                  sqrt(zpl.A * zpl.A + zpl.B*zpl.B + zpl.C*zpl.C);
                // reduce the effect of shading -- we don't want parts of the image
                // to be completely blackened out.
                f := (abs(f)/4) + 0.75;
                if ( (f < 0.74) or (f > 1.0)) then begin
                  // should never happen
                  f := 1.0;
                end;
                Ptr := PByte(pixel + pixel + pixel + integer(zp.DIB.Bits));
                Ptr^ := Trunc(f*zc.b);
                Inc(Ptr);
                Ptr^ := Trunc(f*zc.g);
                Inc(Ptr);
                Ptr^ := Trunc(f*zc.r);
(*
                zp.DIB.SetPixel(p.x, zp.cols-1-p.y, (Trunc(f*zc.r) shl 16) or
                                                  (Trunc(f*zc.g) shl 8) or
                                                  (Trunc(f*zc.b) shl 0));
*)
              end
              else begin
                Ptr := PByte(pixel + pixel + pixel + integer(zp.DIB.Bits));
                Ptr^ := zc.b;
                Inc(Ptr);
                Ptr^ := zc.g;
                Inc(Ptr);
                Ptr^ := zc.r;
(*
                zp.DIB.SetPixel(p.x, zp.cols-1-p.y, (zc.r shl 16) or
                                                  (zc.g shl 8) or
                                                  (zc.b shl 0));
*)
              end;

						end;
					end;
				end;
//			end;
      Inc(p.x);
		end;
    Inc(p.y);
	end;
end;

procedure zAddLine(var zp: ZIMAGE; const pc1, pc2: PZPOINT; const zcp, zcs: PZCOLOR; const replace: integer);
var
	yslope:   MyReal;
	zslope:   MyReal;
	x,y,z:    integer;
	temp:     integer;
	transposed: integer;
	pA, pB:   ZPOINT;
	p1, p2:   PZPOINT;
	pixel:    integer;
	zc:       PZCOLOR;
  x1, x2, y1, y2: integer;
  xxx, yyy: integer;
  mxpixel:  integer;
  xq, yq:   integer;
  Ptr:      PByte;
begin
  transposed := 0;
  mxpixel := zp.Rows*zp.Cols;

	if (not Assigned(zp.zbuffer)) then begin
		Exit;
	end;
	begin
		x1 := maxmin(pc1.x, pc2.x,0);
		x2 := minmax(pc1.x, pc2.x,(zp.rows-1));
		y1 := maxmin(pc1.y, pc2.y,0);
		y2 := minmax(pc1.y, pc2.y,(zp.cols-1));
		zp.dirty_x1 := min(x1, zp.dirty_x1);
		zp.dirty_x2 := max(x2, zp.dirty_x2);
		zp.dirty_y1 := min(y1, zp.dirty_y1);
		zp.dirty_y2 := max(y2, zp.dirty_y2);
//		zp.extent_x1 := min(x1,zp.extent_x1);
//		zp.extent_x2 := max(x2,zp.extent_x2);
//		zp.extent_y1 := min(y1,zp.extent_y1);
//		zp.extent_y2 := max(y2,zp.extent_y2);
//    Debug.Print(Format('add_line: %d %d %d %d',
//        [zp.dirty_x1, zp.dirty_x2, zp.dirty_y1, zp.dirty_y2]));
	end;

	pA := pc1^;
	pB := pc2^;
	p1 := @pA;
	p2 := @pB;

	// Handle vertical line case
	if (p1.x = p2.x) then begin
		if(p1.y > p2.y) then begin
			// swap points
			temp := p1.x;
			p1.x := p2.x;
			p2.x := temp;
			temp:=p1.y;
			p1.y:=p2.y;
			p2.y := temp;
			temp:=p1.z;
			p1.z:=p2.z;
			p2.z := temp;
		end;
		if (p2.y = p1.y) then begin
			zslope := 0.0;
		end
    else begin
			zslope := (p2.z - p1.z)/(p2.y-p1.y);
		end;
		if ((p1.x >=0) and (p1.x < zp.rows)) then begin
			for y:= max(0,p1.y) to min(((zp.cols)-1),p2.y) do begin
				z := Trunc(zslope * (y-p1.y)) + p1.z;
				pixel := p1.x + (zp.cols-1-y)*zp.rows {p1.x+y*zp.rows{p1.x*zp.cols+y};
				if((pixel>=0) and (pixel < mxpixel{(zp.rows * zp.cols)})) then begin
					if (((replace = 0) and (zp.zbuffer[pixel] > z)) or
						((replace = 1) and (zp.zbuffer[pixel] >= z))) then begin
						// dither colors
						if ((Odd(p1.x+y)){ mod 2) <> 0}) then begin
							zc := zcp;
						end
            else begin
							zc := zcs;
						end;
						if (zc.a > 0) then begin
							zp.zbuffer[pixel] := z;
(*
							zp.r[pixel]:= zc.r;
							zp.g[pixel]:= zc.g;
							zp.b[pixel]:= zc.b;
*)
              Ptr := PByte(pixel + pixel + pixel + integer(zp.DIB.Bits));
              Ptr^ := zc.b;
              Inc(Ptr);
              Ptr^ := zc.g;
              Inc(Ptr);
              Ptr^ := zc.r;
(*
              zp.DIB.SetPixel(p1.x, zp.cols-1-y, (zc.r shl 16) or
                                       (zc.g shl 8) or
                                       (zc.b shl 0));
*)
						end;
					end;
				end;
			end;
		end;
		Exit;
	end;
	// Handle horizontal line case
	if (p1.y = p2.y) then begin
		if(p1.x > p2.x) then begin
			// swap points
			temp := p1.x;
			p1.x := p2.x;
			p2.x := temp;
			temp:=p1.y;
			p1.y:=p2.y;
			p2.y := temp;
			temp:=p1.z;
			p1.z:=p2.z;
			p2.z := temp;
		end;
		if (p2.x = p1.x) then begin
			zslope := 0.0;
		end
    else begin
			zslope := (p2.z - p1.z)/(p2.x-p1.x);
		end;
		if ((p1.y >=0) and (p1.y < zp.cols)) then begin
			for x:= max(0,p1.x) to min(((zp.rows)-1),p2.x) do begin
				z := Trunc(zslope * (x-p1.x)) + p1.z;
				pixel := x + (zp.cols-1-p1.y)*zp.rows {x+p1.y*zp.rows{x*zp.cols+p1.y};
				if((pixel>=0) and (pixel < mxpixel{(zp.rows * zp.cols)})) then begin
					if (((replace = 0) and (zp.zbuffer[pixel] > z)) or
						((replace = 1) and (zp.zbuffer[pixel] >= z))) then begin
						// dither colors
						if ((Odd(x+p1.y)){ mod 2) <> 0}) then begin
							zc := zcp;
						end
            else begin
							zc := zcs;
						end;
						if (zc.a > 0) then begin
							zp.zbuffer[pixel] := z;
(*
							zp.r[pixel]:= zc.r;
							zp.g[pixel]:= zc.g;
							zp.b[pixel]:= zc.b;
*)
              Ptr := PByte(pixel + pixel + pixel + integer(zp.DIB.Bits));
              Ptr^ := zc.b;
              Inc(Ptr);
              Ptr^ := zc.g;
              Inc(Ptr);
              Ptr^ := zc.r;
(*
              zp.DIB.SetPixel(x, zp.cols-1-p1.y, (zc.r shl 16) or
                                       (zc.g shl 8) or
                                       (zc.b shl 0));
*)
						end;
					end;
				end;
			end;
		end;
		Exit;
	end;
	// Handle diagonals
	if(p1.x > p2.x) then begin
		// swap points
		temp := p1.x;
		p1.x := p2.x;
		p2.x := temp;
		temp:=p1.y;
		p1.y:=p2.y;
		p2.y := temp;
		temp:=p1.z;
		p1.z:=p2.z;
		p2.z := temp;
	end;
	yslope := (p2.y - p1.y) / (p2.x-p1.x);
	// Transpose if slope too steep.
	if (abs(yslope) > 1.0) then begin
		transposed := 1;
		temp := p1.x;
		p1.x := p1.y;
		p1.y := temp;
		temp := p2.x;
		p2.x := p2.y;
		p2.y := temp;
		if (p1.x > p2.x) then begin
			// swap points
			temp := p1.x;
			p1.x := p2.x;
			p2.x := temp;
			temp:=p1.y;
			p1.y:=p2.y;
			p2.y := temp;
			temp:=p1.z;
			p1.z:=p2.z;
			p2.z := temp;
		end;
		yslope := (p2.y - p1.y) / (p2.x-p1.x);
	end;
	// Handle diagonals with yslope < 1
	zslope := (p2.z - p1.z)/(p2.x-p1.x);
  if transposed = 0 then begin
    xxx := zp.rows;
    yyy := zp.cols;
  end
  else begin
    xxx := zp.cols;
    yyy := zp.rows;
  end;

	for x:= max(0,p1.x) to min(xxx-1,p2.x) do begin
		y := Trunc(yslope * (x-p1.x)) + p1.y;
		if ((y>0) and (y < yyy)) then begin
			z := Trunc(zslope * (x-p1.x)) + p1.z;
			if (transposed = 0) then begin
        xq := x;
        yq := y;
				pixel := x + (zp.cols-1-y)*zp.rows {x+y*zp.rows{x*zp.cols+y};
			end
      else begin
        xq := y;
        yq := x;
				pixel := y + (zp.cols-1-x)*zp.rows {y+x*zp.rows{y*zp.cols+x};
			end;
			if((pixel>=0) and (pixel < mxpixel{(zp.rows * zp.cols)})) then begin
				if (((replace = 0) and (zp.zbuffer[pixel] > z)) or
					((replace = 1) and (zp.zbuffer[pixel] >= z))) then begin
						// dither colors
						if ((Odd(x+y)){ mod 2) <> 0}) then begin
							zc := zcp;
						end
            else begin
							zc := zcs;
						end;
						if (zc.a > 0) then begin
							zp.zbuffer[pixel] := z;
(*
							zp.r[pixel]:= zc.r;
							zp.g[pixel]:= zc.g;
							zp.b[pixel]:= zc.b;
              zp.DIB.SetPixel(pixel div zp.cols, zp.cols-1-pixel mod zp.cols, (zc.r shl 16) or
                                       (zc.g shl 8) or
                                       (zc.b shl 0));
*)
              Ptr := PByte(pixel + pixel + pixel + integer(zp.DIB.Bits));
              Ptr^ := zc.b;
              Inc(Ptr);
              Ptr^ := zc.g;
              Inc(Ptr);
              Ptr^ := zc.r;
(*
              zp.DIB.SetPixel(xq, zp.cols-1-yq, (zc.r shl 16) or
                                       (zc.g shl 8) or
                                       (zc.b shl 0));
*)
						end;
				end;
			end;
		end;
	end;
end;


procedure zDraw(const zDC: pointer);
begin
	if (z.zbuffer <> nil) then begin
		platform_zDraw(z, zDC);
	end;
end;

function zReset(rows, cols: integer): integer;
var
  rc: integer;
begin
	repeat
		rc := init_zimage(z, rows, cols);
		if (rc <> 0) then begin
			if ((rows > 64) and (cols > 64)) then begin
				// Maybe window is too large and there is not
				// enough memory to support it.  Try making the
				// the drawing smaller.
				rows := rows div 2;
				cols := cols div 2;
			end
      else begin
				// Severe memory problems
				Result := -1;
			end;
		end;
  until (rc = 0);
	Result := 0;
end;

procedure zFree;
begin
	if (z.zbuffer <> nil) then begin
		FreeMem(z.zbuffer);
		z.zbuffer := nil;
	end;
(*
	if (z.r <> nil) then begin
		FreeMem(z.r);
		z.r := nil;
	end;
	if (z.g <> nil) then begin
		FreeMem(z.g);
		z.g := nil;
	end;
	if (z.b <> nil) then begin
		FreeMem(z.b);
		z.b := nil;
	end;
*)
  if Assigned(z.DIB) then begin
    z.DIB.Free;
    z.DIB := nil;
  end
end;

procedure zStep(const step, pause: integer);
begin
	platform_step(step, include_stack_ptr, pause, z);
end;

function zGetRowsize: integer;
begin
	Result := z.rows;
end;

function zGetColsize: integer;
begin
	Result := z.cols;
end;

end.
